import numpy as np

TILE_SIZE = 256
BASE_4 = np.array(["0", "1", "2", "3"], dtype="<U1")


def _project_longitude(longitude):
    """
    Projects a vector of longitudes to the interval [0, 1].
    """
    return 0.5 + np.clip(longitude, -180, 180) / 360


def _project_latitude(latitude):
    """
    Projects a vector of latitudes to the interval [0, 0.99999].
    This limits the latitude to the range [-85.05112877, 85.05112877] degrees.
    """
    sin_y = np.sin((np.clip(latitude, -85.05112877, 85.05112877) * np.pi) / 180)
    return 0.5 - np.log((1 + sin_y) / (1 - sin_y)) / (4 * np.pi)


def project(latitude, longitude):
    """
    Projects vectors of latitude and longitude to the interval [0, 0.99999] and [0, 1], respectively.
    """
    return _project_longitude(longitude), _project_latitude(latitude)


################################################ to TILE
##############################################
def _projection_to_tile(x, zoom):
    """
    Returns the tile coordinate based on zoom level for a given projected latitude or longitude.
    """
    return np.int_(x * (1 << zoom))


def lat_lon_to_tile(latitude, longitude, zoom):
    """
    Returns the [tile_x] and [tile_y] coordinates for given vectors of latitude, longitude and zoom level.
    """
    proj_x, proj_y = _project_longitude(longitude), _project_latitude(latitude)
    return _projection_to_tile(proj_x, zoom), _projection_to_tile(proj_y, zoom)


def lat_lon_to_tile_tuple(latitude, longitude, zoom):
    """
    Returns the [[tile_x, tile_y]] coordinates for given vectors of latitude, longitude and zoom level.
    """
    tile_x, tile_y = lat_lon_to_tile(latitude, longitude, zoom)
    return np.array((tile_x, tile_y)).T


def pixel_to_tile(pixel_x, pixel_y):
    """
    Returns the [tile_x] and [tile_y] coordinates for given vectors of pixel_x and pixel_y coordinates.
    """
    return np.int_(pixel_x // TILE_SIZE), np.int_(pixel_y // TILE_SIZE)


def pixel_to_tile_tuple(pixel_x, pixel_y):
    """
    Returns the [[tile_x, tile_y]] coordinates for given vectors of pixel_x and pixel_y coordinates.
    """
    tile_x, tile_y = pixel_to_tile(pixel_x, pixel_y)
    return np.array((tile_x, tile_y)).T


def pixel_tuple_to_tile_tuple(pixel_tuple):
    """
    Returns the [[tile_x, tile_y]] coordinates for given vector of [[pixel_x], [pixel_y]] coordinates.
    """
    return np.int_(pixel_tuple // TILE_SIZE)


def quadkey_to_tile(quadkeys):
    """
    Converts QuadKeys to vectors of [tile_x] and [tile_y] coordinates.
    """
    if isinstance(quadkeys, list):
        quadkeys = np.array(quadkeys, dtype="U").reshape(len(quadkeys), 1)
    elif isinstance(quadkeys, str):
        quadkeys = np.array(quadkeys, dtype="U").reshape(1, 1)

    quadkeys = quadkeys.view("U1").astype(np.int_)
    shift_sequence = np.arange(quadkeys.shape[-1])[::-1]

    to_tile_y = ((quadkeys & 0b10) >> 1) << shift_sequence
    to_tile_x = (quadkeys & 0b01) << shift_sequence
    return to_tile_x.sum(1), to_tile_y.sum(1)


def quadkey_to_tile_tuple(quadkeys):
    """
    Converts QuadKeys to vector of tile tuple [[tile_x, tile_y]] coordinates.
    """
    tile_x, tile_y = quadkey_to_tile(quadkeys)
    return np.array((tile_x, tile_y)).T


##############################################
################################################


################################################ to QuadKey
##############################################
def _broadcast_array(x, repeat):
    """
    Broadcasts array X to a shape of (repeat, len(X))

                                     [[1, 1, 1, 1, 1]
    _broadcast_array([1,5,7], 5) -->  [5, 5, 5, 5, 5]
                                      [7, 7, 7, 7, 7]]
    """
    return np.broadcast_to(x, (repeat, x.size)).T


def _interleave_base4(tile_x_broadcasted, tile_y_broadcasted, zoom):
    """
    To convert tile coordinates into a quadkey, the bits of the Y and X coordinates are interleaved,
    and the result is interpreted as a base-4 number (with leading zeros maintained) and converted into a string.
    For instance, given tile XY coordinates of (3, 5) at level 3, the quadkey is determined as follows:
        tileX = 3 = 0112
        tileY = 5 = 1012
        quadkey = 1001112= 2134= “213”
    """
    shift_sequence = np.arange(zoom)[::-1]
    part_y = (np.right_shift(tile_y_broadcasted, shift_sequence) & 0b1) << 1
    part_x = np.right_shift(tile_x_broadcasted, shift_sequence) & 0b1
    return BASE_4[part_y | part_x].view(f"U{zoom}")


def tile_to_quadkey(tile_x, tile_y, zoom):
    """
    Converts [tile_x] and [tile_y] vectors to [quadkey] for zoom level `zoom`.
    The `zoom` level used here should be the same as the one used to create the tiles.
    """
    tile_x_broadcasted = _broadcast_array(np.int_(tile_x), zoom)
    tile_y_broadcasted = _broadcast_array(np.int_(tile_y), zoom)
    return _interleave_base4(tile_x_broadcasted, tile_y_broadcasted, zoom).flatten()


def lat_lon_to_quadkey(latitude, longitude, zoom):
    """Converts [latitude] and [longitude] vectors to [quadkey] for wanted zoom level."""
    tile_x, tile_y = lat_lon_to_tile(latitude, longitude, zoom)
    return tile_to_quadkey(tile_x, tile_y, zoom)


def pixel_to_quadkey(pixel_x, pixel_y, zoom):
    """Converts [pixel_x] and [pixel_y] vectors to [quadkey]."""
    tile_x, tile_y = pixel_to_tile(pixel_x, pixel_y)
    return tile_to_quadkey(tile_x, tile_y, zoom)


##############################################
################################################


################################################ to PIXEL
##############################################
def _projection_to_pixel(x, zoom):
    """Returns a PIXEL coordinate based on zoom level for projected lat/lon."""
    return np.int_(x * (TILE_SIZE << zoom))


def lat_lon_to_pixel(latitude, longitude, zoom):
    """Returns [pixel_x] and [pixel_y] coords for zoom level."""
    proj_x, proj_y = _project_longitude(longitude), _project_latitude(latitude)
    return _projection_to_pixel(proj_x, zoom), _projection_to_pixel(proj_y, zoom)


def lat_lon_to_pixel_tuple(latitude, longitude, zoom):
    """Returns [(pixel_x, pixel_y)] coords for zoom level."""
    pixel_x, pixel_y = lat_lon_to_pixel(latitude, longitude, zoom)
    return np.array((pixel_x, pixel_y)).T


def tile_to_pixel(tile_x, tile_y):
    """Convert [tile_x] and [tile_y] vectors to [[min_x, min_y, max_x, max_y]] pixel coordinates."""
    return np.array(
        (
            tile_x * TILE_SIZE,
            tile_y * TILE_SIZE,
            (tile_x + 1) * TILE_SIZE,
            (tile_y + 1) * TILE_SIZE,
        )
    ).T


def quadkey_to_pixel(quadkeys):
    """Convert [quadkey] vector to [[min_x, min_y, max_x, max_y]]  pixel coordinates."""
    tile_x, tile_y = quadkey_to_tile(quadkeys)
    return tile_to_pixel(tile_x, tile_y)


##############################################
################################################


################################################ to LatLon
##############################################
def _pixel_to_projection(x, zoom):
    """Transforms pixel coordinate to range [0, 1]."""
    max_pixel = TILE_SIZE << zoom
    return np.clip(x, 0, max_pixel) / max_pixel


def _tile_to_projection(x, zoom):
    """Transforms tile to projection."""
    return x / (1 << zoom)


def _projection_to_latitude(y):
    """Returns latitude from projection value."""
    return 90 - 360 * np.arctan(np.exp((y - 0.5) * 2 * np.pi)) / np.pi


def _projection_to_longitude(x):
    """Returns longitude from projection value."""
    return 360 * (x - 0.5)


def _tile_to_latitude(x, zoom):
    """Gets latitude of top left corner of tile."""
    return _projection_to_latitude(_tile_to_projection(x, zoom))


def _tile_to_longitude(x, zoom):
    """Gets latitude of top left corner of tile."""
    return _projection_to_longitude(_tile_to_projection(x, zoom))


def tile_to_lat_lon(tile_x, tile_y, zoom):
    """Convert [tile_x] and [tile_y] vectors to [[min_lon, min_lat, max_lon, max_lat]]."""
    min_lon = _tile_to_longitude(tile_x, zoom)
    max_lat = _tile_to_latitude(tile_y, zoom)
    max_lon = _tile_to_longitude(tile_x + 1, zoom)
    min_lat = _tile_to_latitude(tile_y + 1, zoom)
    return np.array((min_lon, min_lat, max_lon, max_lat)).T.reshape(-1, 4)


def quadkey_to_lat_lon(quadkeys):
    """Convert [quadkey] vector to [[min_lon, min_lat, max_lon, max_lat]]."""
    zoom = len(quadkeys) if isinstance(quadkeys, str) else len(quadkeys[0])
    tile_x, tile_y = quadkey_to_tile(quadkeys)
    return tile_to_lat_lon(tile_x, tile_y, zoom)


def pixel_to_lat_lon(pixel_x, pixel_y, zoom):
    """Converts [pixel_x] and [pixel_y] vectors to [latitude] and [longitude] for zoom level."""
    proj_x, proj_y = _pixel_to_projection(pixel_x, zoom), _pixel_to_projection(pixel_y, zoom)
    return _projection_to_latitude(proj_y), _projection_to_longitude(proj_x)


def pixel_to_lat_lon_tuple(pixel_x, pixel_y, zoom):
    """Converts [pixel_x] and [pixel_y] vectors to [[latitude, longitude]] for zoom level."""
    lat, lon = pixel_to_lat_lon(pixel_x, pixel_y, zoom)
    return np.array((lat, lon)).T


def pixel_tuple_to_lat_lon_tuple(latitude, longitude, zoom):
    "TODO"


##############################################
################################################


################################################ boundaries
##############################################
def lat_lon_bounds_to_tile_range(bounds, zoom):
    """Converts bounds [[min_lon, min_lat, max_lon, max_lat]] from GeoDataFrame to tile coords [[min_x, min_y, max_x, max_y]]."""
    lon_min, lat_min, lon_max, lat_max = bounds.T
    tile_x_min, tile_y_max = lat_lon_to_tile(lat_max, lon_min, zoom)
    tile_x_max, tile_y_min = lat_lon_to_tile(lat_min, lon_max, zoom)
    return np.array([tile_x_min, tile_y_min, tile_x_max, tile_y_max]).T



"""
A module for vectorized geohashing in Python.
Vectorized functions that allow efficient transformation of latitude and longitude coordinates into Map Tile
coordinates, Map Pixel coordinates or QuadKeys. This can be particularly useful in (GIS) geographic information systems
and mapping applications. By utilizing vectorized functions, transformation can be performed on large datasets,
with minimal impact on performance. Additionally, only using numpy as dependency these functions can be easily
incorporated into existing code and workflows.

Platform Support:
    This module is supported on Unix and Windows platforms.

Module Author:
    name : Filip Jakovljevic
    email : fillix96@gmail.com
"""
try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:
    from importlib_metadata import PackageNotFoundError, version

from .vec_geohash import *  # noqa: F403

try:
    __version__ = version("vec_geohash")
except PackageNotFoundError:
    # If the package is not installed, don't add __version__
    pass

__author__ = "FJakovljevic"



import unittest

import numpy as np
import vec_geohash


class TestLatLon(unittest.TestCase):
    def test_latlon_single(self):
        # test input values
        lat = 53.1231276599
        lon = 82.6978699112
        zoom = 9

        # expected output for tile tuple
        expected_output = (373, 166)
        self.assertEqual(vec_geohash.lat_lon_to_tile(lat, lon, zoom), expected_output)

        # expected output for quadkey
        expected_output = np.array(["121310321"], dtype="<U9")
        np.testing.assert_array_equal(
            vec_geohash.lat_lon_to_quadkey(lat, lon, zoom), expected_output
        )

        # expected output for pixel tuple
        expected_output = (95645, 42622)
        np.testing.assert_array_equal(
            vec_geohash.lat_lon_to_pixel(lat, lon, zoom), expected_output
        )

    def test_latlon_vector(self):
        # test input values
        lat = np.array([53.1231276599, 41.85])
        lon = np.array([82.6978699112, -87.65])
        zoom = 9

        # expected output for tile tuple
        expected_output = np.array([[373, 166], [131, 190]])
        np.testing.assert_array_equal(
            vec_geohash.lat_lon_to_tile_tuple(lat, lon, zoom), expected_output
        )

        # expected output for quadkey
        expected_output = np.array(["121310321", "030222231"], dtype="<U9")
        np.testing.assert_array_equal(
            vec_geohash.lat_lon_to_quadkey(lat, lon, zoom), expected_output
        )

        # expected output for pixel tuple
        expected_output = np.array([[95645, 42622], [33623, 48729]])
        np.testing.assert_array_equal(
            vec_geohash.lat_lon_to_pixel_tuple(lat, lon, zoom), expected_output
        )


class TestPixel(unittest.TestCase):
    def test_pixel_single(self):
        # test input values
        lat = 95645
        lon = 42622
        zoom = 9

        # expected output for tile tuple
        expected_output = (373, 166)
        self.assertEqual(vec_geohash.pixel_to_tile(lat, lon), expected_output)

        # expected output for quadkey
        expected_output = np.array(["121310321"], dtype="<U9")
        np.testing.assert_array_equal(
            vec_geohash.pixel_to_quadkey(lat, lon, zoom), expected_output
        )

        # expected output for latlon tuple
        expected_output = (53.1231276599, 82.6978699112)
        np.testing.assert_almost_equal(
            vec_geohash.pixel_to_lat_lon(lat, lon, zoom), expected_output, decimal=3
        )

    def test_pixel_vector(self):
        # test input values
        lat = np.array([95645, 33623])
        lon = np.array([42622, 48729])
        zoom = 9

        # expected output for tile tuple
        expected_output = np.array([[373, 166], [131, 190]])
        np.testing.assert_array_equal(
            vec_geohash.pixel_to_tile_tuple(lat, lon), expected_output
        )

        # expected output for quadkey
        expected_output = np.array(["121310321", "030222231"], dtype="<U9")
        np.testing.assert_array_equal(
            vec_geohash.pixel_to_quadkey(lat, lon, zoom), expected_output
        )

        # expected output for latlon tuple
        expected_output = np.array([[53.1231276599, 82.6978699112], [41.851, -87.652]])
        np.testing.assert_almost_equal(
            vec_geohash.pixel_to_lat_lon_tuple(lat, lon, zoom),
            expected_output,
            decimal=3,
        )


class TestTile(unittest.TestCase):
    def test_tile_single(self):
        # test input values
        tile_x = 1
        tile_y = 1
        zoom = 1

        # expected output for quadkey
        expected_output = np.array(["3"], dtype="<U1")
        np.testing.assert_array_equal(
            vec_geohash.tile_to_quadkey(tile_x, tile_y, zoom), expected_output
        )

        # expected output for latlon
        expected_output = np.array([[0.0, -85.05112878, 180.0, 0.0]])
        np.testing.assert_almost_equal(
            vec_geohash.tile_to_lat_lon(tile_x, tile_y, zoom), expected_output
        )

        # expected output for pixel
        expected_output = np.array([256, 256, 512, 512])
        np.testing.assert_array_equal(
            vec_geohash.tile_to_pixel(tile_x, tile_y), expected_output
        )

    def test_tile_vec(self):
        # test input values
        tile_x = np.array([0, 3])
        tile_y = np.array([0, 3])
        zoom = 2

        # expected output for quadkey
        expected_output = np.array(["00", "33"], dtype="<U2")
        np.testing.assert_array_equal(
            vec_geohash.tile_to_quadkey(tile_x, tile_y, zoom), expected_output
        )

        # expected output for latlon
        expected_output = np.array(
            [
                [-180.0, 66.5132604, -90.0, 85.0511288],
                [90.0, -85.0511288, 180.0, -66.5132604],
            ]
        )
        np.testing.assert_almost_equal(
            vec_geohash.tile_to_lat_lon(tile_x, tile_y, zoom), expected_output
        )

        # expected output for pixel
        expected_output = np.array([[0, 0, 256, 256], [768, 768, 1024, 1024]])
        np.testing.assert_array_equal(
            vec_geohash.tile_to_pixel(tile_x, tile_y), expected_output
        )


class TestQuadkey(unittest.TestCase):
    def test_quadkey_single(self):
        quadkey = "121310321"

        # expected output for tile tuple
        expected_output = np.array([[373, 166]])
        np.testing.assert_array_equal(
            vec_geohash.quadkey_to_tile_tuple(quadkey), expected_output
        )

        # expected output for latlon
        expected_output = np.array([[82.265625, 52.908902, 82.96875, 53.330873]])
        np.testing.assert_almost_equal(
            vec_geohash.quadkey_to_lat_lon(quadkey), expected_output
        )

        # expected output for pixel
        expected_output = np.array([[95488, 42496, 95744, 42752]])
        np.testing.assert_almost_equal(
            vec_geohash.quadkey_to_pixel(quadkey), expected_output
        )

    def test_quadkey_vector(self):
        quadkey = ["121310321", "030222231"]

        # expected output for tile tuple
        expected_output = np.array([[373, 166], [131, 190]])
        np.testing.assert_array_equal(
            vec_geohash.quadkey_to_tile_tuple(quadkey), expected_output
        )

        # expected output for latlon
        expected_output = np.array(
            [
                [82.2656, 52.9089, 82.9687, 53.3308],
                [-87.8906, 41.5085, -87.1875, 42.0329],
            ]
        )
        np.testing.assert_almost_equal(
            vec_geohash.quadkey_to_lat_lon(quadkey), expected_output, decimal=3
        )

        # expected output for pixel
        expected_output = np.array(
            [[95488, 42496, 95744, 42752], [33536, 48640, 33792, 48896]]
        )
        np.testing.assert_almost_equal(
            vec_geohash.quadkey_to_pixel(quadkey), expected_output, decimal=3
        )


class TestProjection(unittest.TestCase):
    def test_project(self):
        # Test with basic inputs
        longitudes = np.array([0, 45, 90, 135, 180, -45, -90, -135, -180])
        expected_longitudes = np.array(
            [0.5, 0.625, 0.75, 0.875, 1, 0.375, 0.25, 0.125, 0]
        )

        latitudes = np.array([0, 30, 60, -30, -60])
        expected_latitudes = np.array([0.5, 0.4125752, 0.2903996, 0.5874248, 0.7096004])

        projected_longitudes, projected_latitudes = vec_geohash.project(
            latitudes, longitudes
        )
        np.testing.assert_almost_equal(projected_longitudes, expected_longitudes)
        np.testing.assert_almost_equal(projected_latitudes, expected_latitudes)

        # Test with single input value
        latitudes = 0
        longitudes = 0
        expected_longitudes = 0.5
        expected_latitudes = 0.5
        projected_longitudes, projected_latitudes = vec_geohash.project(
            latitudes, longitudes
        )
        np.testing.assert_approx_equal(projected_longitudes, expected_longitudes)
        np.testing.assert_approx_equal(projected_latitudes, expected_latitudes)

        # Test with input values exceeding limits
        latitudes = np.array([100, -100])
        expected_latitudes = np.array([0, 1])

        longitudes = np.array([190, -190])
        expected_longitudes = np.array([1, 0])

        projected_longitudes, projected_latitudes = vec_geohash.project(
            latitudes, longitudes
        )
        np.testing.assert_almost_equal(projected_longitudes, expected_longitudes)
        np.testing.assert_almost_equal(projected_latitudes, expected_latitudes)
